''' 
    Módulo específico para geração de Notas Fiscais de Produto.
    
    Para utilizar esse módulo basta importá-lo da seguinte forma:  
    `from pysisnoapi import nfe`
'''

# =====================================================================
from . import *

import requests
import json
import jsonpickle

from datetime import datetime

# =====================================================================
CSV_HEADERS = [
    'CNPJ EMITENTE',
    'MODELO',
    'CPF/CNPJ DESTINATÁRIO',
    'NOME/RAZAO SOCIAL DESTINATÁRIO',
    'INDICADOR IE',
    'CONSUMIDOR FINAL',
    'FAZ RETENÇÃO DE IMPOSTOS',
    'CEP',
    'BAIRRO',
    'LOGRADOURO',
    'NUMERO',
    'EMAIL DESTINATÁRIO',
    'OPERAÇÃO',
    'FINALIDADE',
    'PRODUTOS',
    'MEIO PAGAMENTO',
    'INFORMAÇÕES COMPLEMENTARES',
    'SIGLA UF',
    'IBGE MUNICÍPIO',
    'COMPLEMENTO ENDEREÇO'
]

BANDEIRAS = {
    '01': 'Visa / Visa Electron',
    '02': 'Mastercard / Maestro',
    '03': 'American Express',
    '04': 'Sorocred',
    '05': 'Diners Club',
    '06': 'Elo',
    '07': 'Hipercard',
    '08': 'Aura',
    '09': 'Cabal',
    '99': 'Outros',
}

CONDICOES_CHASSI = {
    'N': 'Normal',
    'R': 'Remarcado',
}

CONDICOES_VEICULO = {
    '1': 'Acabado',
    '2': 'Inacabado',
    '3': 'Semi-acabado',
}

CORES = {
    '01': 'Amarela',
    '02': 'Azul',
    '03': 'Bege',
    '04': 'Branca',
    '05': 'Cinza',
    '06': 'Dourada',
    '07': 'Grená',
    '08': 'Laranja',
    '09': 'Marrom',
    '10': 'Prata',
    '11': 'Preta',
    '12': 'Rosa',
    '13': 'Roxa',
    '14': 'Verde',
    '15': 'Vermelha',
    '16': 'Fantasia',
}

ESPECIES_VEICULO = {
    '1': 'Passageiro',
    '2': 'Carga',
    '3': 'Misto',
    '4': 'Corrida',
    '5': 'Tração',
    '6': 'Especial',
    '7': 'Coleção',
}

FINALIDADES = {
    '1': 'Normal',
    '2': 'Complementar',
    '3': 'Ajuste',
    '4': 'Devolução ou Retorno',
}

INDICATIVOS_ESCALA = {
    'S': 'Produzido em escala relevante',
    'N': 'Produzido em escala não relevante',
} # TODO: Sugerir utilização de Boolean.

MODALIDADES_FRETE = {
    '9': 'Sem ocorrência de transporte',
    '0': 'Contratação do frete por conta do remetente (CIF)',
    '1': 'Contratação do frete por conta do destinatário (FOB)',
    '2': 'Contratação do frete por conta de terceiros',
    '3': 'Transporte próprio por conta do remetente',
    '4': 'Transporte próprio por conta do destinatário',
}

MODELOS = {
    '55': 'NF-e',
    '65': 'NFC-e',
}

OPERACOES = {
    '0': 'Entrada',
    '1': 'Saída',
}

ORIGENS = {
    '0': 'Nacional',
    '1': 'Estrangeira importação direta',
    '2': 'Estrangeira adquirida mercado interno',
    '3': 'Nacional mercadoria ou bem conteúdo importação superior 40 P',
    '4': 'Nacional produção em conformidade com processos produtivos básicos',
    '5': 'Nacional mercadoria ou bem conteúdo importação inferior 40 P',
    '6': 'Estrangeira importação direta sem similar nacional constante em lista Camex',
    '7': 'Estrangeira adquirida mercado interno sem similar nacional constante em lista Camex',
    '8': 'Nacional mercadoria ou bem conteúdo importação superior 70 P',
}

PRESENCAS = {
    '0': 'Não se aplica',
    '1': 'Operação presencial',
    '2': 'Operação não presencial - Internet',
    '3': 'Operação não presencial - Teleatendimento',
    '4': 'NFC-e em operação com entrega a domicílio',
    '5': 'Operação presencial fora do estabelecimento',
    '9': 'Operação não presencial - Outros',
}

RESTRICOES = {
    '0': 'Não Há',
    '1': 'Alienação Fiduciária',
    '2': 'Arrendamento Mercantil',
    '3': 'Reserva Domínio',
    '4': 'Penhor',
    '9': 'Outras',
}

TIPOS_COMBUSTIVEL = {
    '01': 'Álcool',
    '02': 'Gasolina',
    '03': 'Diesel',
    '04': 'Gasogênio',
    '05': 'Gás Metano',
    '06': 'Elétrico (Fonte Interna)',
    '07': 'Elétrico (Fonte Externa)',
    '08': 'Gasol./Gás Natural/Combustível',
    '09': 'Álcool/Gás Natural',
    '10': 'Diesel/Gás Natural',
    '11': 'Vide Campo Observação',
    '12': 'Álcool/Gás Natural Veicular',
    '13': 'Gasolina/Gás Natural Veicular',
    '14': 'Diesel/Gás Natural Veicular',
    '15': 'Gás Natural Veicular',
    '16': 'Álcool/Gasolina',
    '17': 'Gasolina/Álcool/Gás natural',
    '18': 'Gasolina/Elétrico',
}

TIPOS_EMISSAO = {
    '1': 'Normal',
    '6': 'Contigência SNC-AN',
    '7': 'Contigência SVC-R',
}

TIPOS_OPERACAO = {
    '0': 'Outros',
    '1': 'Venda concessionária',
    '2': 'Faturamento direto consumidor final',
    '3': 'Venda direta grandes consumidores',
}

TIPOS_VEICULO = {
    '06': 'Automóvel',
    '14': 'Caminhão',
    '13': 'Caminhoneta',
    '24': 'Carga',
    '02': 'Ciclomotor',
    '22': 'Especial Ônibus',
    '07': 'Microônibus',
    '23': 'Misto',
    '04': 'Motociclo',
    '03': 'Motoneta',
    '08': 'Ônibus',
    '10': 'Reboque',
    '05': 'Triciclo',
    '17': 'Trator',
}

# =====================================================================
@dataclass
class ObjetoEmissaoNFe(BaseClass):
    numero_nota_sequencial  : str
    serie                   : str
    operacao                : str
    natureza_operacao       : str
    modelo                  : str
    finalidade              : str
    ambiente                : str
    cliente                 : Cliente
    produtos                : List['Produto']
    pedido                  : 'Pedido'
    data_entrada_saida      : datetime
    data_emissao            : datetime
    
    numero_pedido           : Optional[str]                       = None
    transporte              : Optional['Transporte']              = None
    fatura                  : Optional['Fatura']                  = None
    parcelas                : Optional[List['Parcela']]           = None
    exportacao              : Optional['Exportacao']              = None
    nfe_referenciada        : Optional[List[str]]                 = None
    retirada                : Optional['Local']                   = None
    entrega                 : Optional['Local']                   = None
    compra                  : Optional['Compra']                  = None
    indicador_intermediador : Optional[str]                       = None
    informacao_intermediador: Optional['InformacaoIntermediador'] = None
    
    def __post_init__(self):
        self.validate_numero_nota_sequencial()
        self.validate_operacao()
        self.validate_modelo()
        self.validate_finalidade()
        self.validate_ambiente()
    
    def validate_numero_nota_sequencial(self):
        if not isinstance(self.numero_nota_sequencial, str):
            raise TypeError(f'numero_nota_sequencialdeve ser string.')
        if len(self.numero_nota_sequencial) < 1 or len(self.numero_nota_sequencial) > 9:
            raise ValueError(f'numero_nota_sequencial deve conter entre 1 e 9 caracteres.')
    
    def validate_operacao(self):
        if self.operacao not in OPERACOES:
            raise ValueError(f'Operação {self.operacao} inválida')
    
    def validate_modelo(self):
        if self.modelo not in MODELOS:
            raise ValueError(f'Modelo {self.modelo} inválido')
    
    def validate_finalidade(self):
        if self.finalidade not in FINALIDADES:
            raise ValueError(f'Finalidade {self.finalidade} inválida')
    
    def validate_ambiente(self):
        if self.ambiente not in AMBIENTES:
            raise ValueError(f'Ambiente {self.ambiente} inválido')

@dataclass
class Pagamento:
    formas_pagamento: List['FormaPagamento']

@dataclass
class FormaPagamento:
    forma_pagamento         : str
    meio_pagamento          : str
    valor_pagamento         : str
    
    cnpj_credenciadora      : Optional[str] = None
    bandeira                : Optional[str] = None
    autorizacao             : Optional[str] = None
    data_vencimento         : Optional[str] = None
    descricao_meio_pagamento: Optional[str] = None
    # tipo_integracao         : Optional[str] = None    # TODO: Não está na documentação mas é obrigatório caso meio de pagamento seja 'crédito' ou 'débito'. O valor deve ser '1' ou '2'.
    
    def __post_init__(self):
        self.validate_forma_pagamento()
        self.validate_meios_pagamento()
        self.validate_bandeira()
        self.validate_descricao_meio_pagamento()
    
    def validate_forma_pagamento(self, *args, **kwargs):
        if self.forma_pagamento not in FORMAS_PAGAMENTO:
            raise ValueError(f'Forma de pagamento {self.forma_pagamento} inválido')

    def validate_meios_pagamento(self, *args, **kwargs):
        if self.meio_pagamento not in MEIOS_PAGAMENTO:
            raise ValueError(f'Meio de pagamento {self.meio_pagamento} inválido')

    def validate_bandeira(self, *args, **kwargs):
        if self.bandeira:
            if self.bandeira not in BANDEIRAS:
                raise ValueError(f'Bandeira {self.bandeira} inválida.')

    def validate_descricao_meio_pagamento(self, *args, **kwargs):
        if self.meio_pagamento == '99':
            if self.descricao_meio_pagamento is None:
                raise ValueError(f'Necessário preencher "descricao_meio_pagamento" quando a forma de pagamento é "(99) Outros".')
            if len(self.descricao_meio_pagamento) < 2 or len(self.descricao_meio_pagamento) > 60:
                raise ValueError(f'O campo "descricao_meio_pagamento" deve conter entre 2 a 60 caracteres')

@dataclass
class Produto:
    '''Classe Produto

    Raises:
        ValueError: Caso o valor de algum atributo seja inválido.
        TypeError: Caso algum campo obrigatório não for informado.
    '''
    item                            : str   # Número incremental na lista de produtos
    
    cfop                            : str
    nome                            : str
    codigo                          : str
    ncm                             : str
    quantidade                      : str
    unidade                         : str
    subtotal                        : str
    total                           : str
    impostos                        : Impostos
    
    numero_pedido                   : Optional[str]                    = None
    excessao_ibpt                   : Optional[str]                    = None
    peso_liquido                    : Optional[str]                    = None
    peso_bruto                      : Optional[str]                    = None
    origem                          : Optional[str]                    = None
    veiculo_usado                   : Optional[str]                    = None
    ind_escala                      : Optional[str]                    = None
    cnpj_fabricante                 : Optional[str]                    = None
    beneficio_fiscal                : Optional[str]                    = None
    gtin                            : Optional[str]                    = None
    gtin_tributavel                 : Optional[str]                    = None
    cest                            : Optional[str]                    = None
    nve                             : Optional[str]                    = None
    informacoes_adicionais          : Optional[str]                    = None
    veiculo_novo                    : Optional['VeiculoNovo']          = None
    exportacao                      : Optional['ExportacaoIndividual'] = None
    importacao                      : Optional['ImportacaoIndividual'] = None
    rastro                          : Optional['Rastreamento']         = None
    ex_tipi                         : Optional[str]                    = None
    valor_frete                     : Optional[str]                    = None
    valor_seguro                    : Optional[str]                    = None
    valor_desconto                  : Optional[str]                    = None
    valor_outras_despesas_acessorias: Optional[str]                    = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_unidade()
        self.validate_impostos()
        self.validate_origem()
        self.validate_total()
        self.validate_ind_escala()
        
    def validate_unidade(self, *args, **kwargs):
        if self.unidade not in UNIDADES:
            raise ValueError(f'Unidade {self.unidade} inválida.')
        
    def validate_impostos(self, *args, **kwargs):
        if not isinstance(self.impostos, Impostos):
            raise TypeError(f'Campo "impostos" deve ser {repr(Impostos)} e não {type(self.impostos)}')

    def validate_origem(self, *args, **kwargs):
        if self.origem:
            if self.origem not in ORIGENS:
                raise ValueError(f'Origem {self.origem} inválida.')
        
    def validate_total(self, *args, **kwargs):
        quantidade = float(self.quantidade)
        subtotal   = float(self.subtotal)
        total      = float(self.total)
        
        if f'{total:.2f}' != f'{quantidade*subtotal:.2f}':
            raise ValueError(f'Total {total:.2f} é diferente de quantidade {quantidade:.2f} * subtotal {subtotal:.2f} = {quantidade*subtotal:.2f}')
        
    def validate_ind_escala(self, *args, **kwargs):
        if self.ind_escala:
            if self.ind_escala not in INDICATIVOS_ESCALA:
                raise ValueError(f'Indicativo de escala {self.ind_escala} inválido.')
    
@dataclass
class ImpostosProduto(Impostos):
    icms: 'Icms'
    ipi : 'Ipi'

@dataclass
class Pedido:
    presenca                  : str
    pagamento                 : 'Pagamento'
    
    modalidade_frete          : Optional[str] = None
    frete                     : Optional[str] = None
    desconto                  : Optional[str] = None
    total                     : Optional[str] = None
    despesas_acessorias       : Optional[str] = None
    despesas_aduaneiras       : Optional[str] = None
    informacoes_complementares: Optional[str] = None
    informacoes_fisco         : Optional[str] = None
    observacoes_fisco         : Optional[str] = None
    observacoes_contribuinte  : Optional[str] = None
    
    def __post_init__(self):
        self.validate_presenca()
        self.validate_modalidade_frete()
    
    def validate_presenca(self, *args, **kwargs):
        if self.presenca not in PRESENCAS:
            raise ValueError(f'Presença {self.presenca} inválida.')
        
    def validate_modalidade_frete(self, *args, **kwargs):
        if self.modalidade_frete:
            if self.modalidade_frete not in MODALIDADES_FRETE:
                raise ValueError(f'Modalidade de Frete {self.modalidade_frete} inválida.')

@dataclass
class VeiculoNovo:
    tipo_operacao   : str
    chassi          : str
    cor             : str
    cor_descricao   : str
    cilindrada      : str
    peso_liquido    : str
    peso_bruto      : str
    serie           : str
    tipo_combustivel: str
    numero_motor    : str
    dist            : str
    ano_modelo      : str
    ano_fabricacao  : str
    tipo_pintura    : str
    tipo_veiculo    : str
    especie_veiculo : str
    condicao_veiculo: str
    marca_modelo    : str
    cor_denatran    : str
    lotacao         : str
    restricao       : str

    capacidade_maxima_tracao: Optional[str] = None
    condicao_chassi         : Optional[str] = None
    potencia                : Optional[str] = None
   
    def __post_init__(self):
       self.validate_tipo_operacao()
       self.validate_cor()
       self.validate_tipo_combustivel()
       self.validate_tipo_veiculo()
       self.validate_especie_veiculo()
       self.validate_condicao_veiculo()
       self.validate_restricao()
       self.validate_condicao_chassi()
    
    def validate_tipo_operacao(self, *args, **kwargs):
        if self.tipo_operacao not in TIPOS_OPERACAO:
            raise ValueError(f'Tipo de operação {self.tipo_operacao} inválido.')
    
    def validate_cor(self, *args, **kwargs):
        if self.cor not in CORES:
            raise ValueError(f'Cor {self.cor} inválida.')
    
    def validate_tipo_combustivel(self, *args, **kwargs):
        if self.tipo_combustivel not in TIPOS_COMBUSTIVEL:
            raise ValueError(f'Tipo combustível {self.tipo_combustivel} inválido.')
    
    def validate_tipo_veiculo(self, *args, **kwargs):
        if self.tipo_veiculo not in TIPOS_VEICULO:
            raise ValueError(f'Tipo de Veículo {self.tipo_veiculo} inválido.')
    
    def validate_especie_veiculo(self, *args, **kwargs):
        if self.especie_veiculo not in ESPECIES_VEICULO:
            raise ValueError(f'Espécie de Veículo {self.especie_veiculo} inválida.')
    
    def validate_condicao_veiculo(self, *args, **kwargs):
        if self.condicao_veiculo not in CONDICOES_VEICULO:
            raise ValueError(f'Condição do véiculo {self.condicao_veiculo} inválida.')
    
    def validate_restricao(self, *args, **kwargs):
        if self.restricao not in RESTRICOES:
            raise ValueError(f'Restrição {self.restricao} inválida.')
    
    def validate_condicao_chassi(self, *args, **kwargs):
        if self.condicao_chassi:
            if self.condicao_chassi not in CONDICOES_CHASSI:
                raise ValueError(f'Condição do Chassi {self.condicao_chassi} inválida.')

@dataclass
class ExportacaoIndividual:
    drawback      : Optional[str] = None
    reg_exportacao: Optional[str] = None
    nfe_exportacao: Optional[str] = None
    qtd_exportacao: Optional[str] = None

@dataclass
class ImportacaoIndividual:
    numero_registro                   : Optional[str]                                = None
    data_registro                     : Optional[str]                                = None
    cod_exportador                    : Optional[str]                                = None
    cod_via_transporte_internacional  : Optional[str]                                = None
    valor_afrmm                       : Optional[str]                                = None
    cod_forma_importacao_intermediacao: Optional[str]                                = None
    uf_desembaraco                    : Optional[str]                                = None
    local_desembaraco                 : Optional[str]                                = None
    data_desembaraco                  : Optional[str]                                = None
    cnpj_terceiro                     : Optional[str]                                = None
    uf_terceiro                       : Optional[str]                                = None
    iof                               : Optional[str]                                = None
    valor_despesas_aduaneiras         : Optional[str]                                = None
    adicoes                           : Optional[List['DeclaracaoImportacaoAdicao']] = None

@dataclass
class Rastreamento:
    lote           : str
    quantidade     : str
    data_fabricacao: str
    data_validade  : str
    
    codigo_agregacao: Optional[str] = None

@dataclass
class Fatura:
    numero       : str
    valor        : str
    desconto     : str
    valor_liquido: str

@dataclass
class Parcela:
    vencimento: str
    valor     : str

@dataclass
class Exportacao:
    uf_embarque   : str
    local_embarque: str
    
    local_despacho: Optional[str] = None

@dataclass
class Local:
    uf                  : str
    codigo_municipio    : str
    descricao_municipio : str
    bairro              : str
    logradouro          : str
    numero              : str
    complemento         : str
    
    cpf : Optional[str] = None
    cnpj: Optional[str] = None

@dataclass
class Compra:
    contrato    : Optional[str] = None
    nota_empenho: Optional[str] = None
    pedido      : Optional[str] = None

@dataclass
class InformacaoIntermediador:
    cnpj                                : Optional[str] = None
    identificador_cadastro_intermediador: Optional[str] = None

@dataclass
class PaginaNotas:
    total           : Optional[str]                = None
    itens_por_pagina: Optional[str]                = None
    pagina_atual    : Optional[str]                = None
    itens           : Optional[List['NotaFiscal']] = None

# =====================================================================
def emitir(token_emissor: str, 
           token_secret_emissor: str,
           token_empresa: str, 
           token_secret_empresa: str,
           objetoNfe: ObjetoEmissaoNFe,
           tipo_emissao: str,
           *args, **kwargs):
    '''Endpoint utilizado para efetivamente emitir uma nota fiscal eletrônica.

    Args:
        nfe (ObjetoEmissaoNFe): Nota Fiscal a ser emitida.
        
        tipo_emissao (str): Tipo de Emissão  
            1: Normal  
            6: Contigência SNC-AN  
            7: Contigência SVC-RS  
    '''

    headers = HEADERS.copy()
    
    # -----------------------------------------
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    validate_tokens(token_empresa, token_secret_empresa)
    headers['token-empresa']        = token_empresa
    headers['token-secret-empresa'] = token_secret_empresa
    
    # -----------------------------------------
    if not tipo_emissao or tipo_emissao not in TIPOS_EMISSAO:
        raise ValueError(f'Tipo de Emissão {tipo_emissao} inválido')
    
    headers['tipo-emissao'] = tipo_emissao
    
    # -----------------------------------------
    json_str = jsonpickle.encode(objetoNfe.as_filtered_dict(), unpicklable=False)
    
    url = f'{BASE_URL}/nfe'
    response = requests.post(url, headers=headers, json=json.loads(json_str))
        
    match response.status_code:
        case 200:
            return response.json()
        case 412:
            return response.json()
        case _:
            return response.text
    
def corrigir(token_emissor: str, 
             token_secret_emissor: str, 
             token_empresa:str, 
             token_secret_empresa:str, 
             *args, **kwargs):
    raise NotImplementedError

def cancelar(token_emissor: str, 
             token_secret_emissor: str,
             token_empresa:str, 
             token_secret_empresa:str, 
             *args, **kwargs):
    raise NotImplementedError

def validar(token_emissor: str, 
            token_secret_emissor: str,
            token_empresa:str, 
            token_secret_empresa:str, 
            objetoNfe:ObjetoEmissaoNFe, 
            tipo_emissao:str,
            *args, **kwargs) -> str:
    '''Endpoint utilizado para validar a nota fiscal eletrônica antes de emitir.

    Args:
        nfe (ObjetoEmissaoNFe): Nota Fiscal a ser validada.
        
        tipo_emissao (str): Tipo de Emissão  
            1: Normal  
            6: Contigência SNC-AN  
            7: Contigência SVC-RS  
    '''
    
    headers = HEADERS.copy()
    
    # -----------------------------------------
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    validate_tokens(token_empresa, token_secret_empresa)
    headers['token-empresa']        = token_empresa
    headers['token-secret-empresa'] = token_secret_empresa
    
    # -----------------------------------------
    if not tipo_emissao or tipo_emissao not in TIPOS_EMISSAO:
        raise ValueError(f'Tipo de Emissão {tipo_emissao} inválido')
    
    headers['tipo-emissao'] = tipo_emissao
    json_str = jsonpickle.encode(objetoNfe.as_filtered_dict(), unpicklable=False)
    
    url = f'{BASE_URL}/nfe/validacao-nota'
    response = requests.post(url, headers=headers, json=json.loads(json_str))
    
    match response.status_code:
        case 200:
            return response.json()
        case _:
            return response.text

def listar(token_emissor: str, 
           token_secret_emissor: str,
           qtd:str = None, 
           pagina:str = None, 
           *args, **kwargs) -> List[NotaFiscal]:
    '''Recupera as notas fiscais.
    
    No Distrito Federal (DF), antes de 01/2023, as notas fiscais de serviço eram emitidas como NFe.

    Args:
        qtd (str, optional): Quantidade de notas por página.
        pagina (str, optional): Página a ser retornada.

    Returns:
        List[NotaFiscal]: Lista contendo as notas fiscais.
    '''
    headers = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    params   = {}
    
    if qtd:
        params['qtd'] = qtd
    if pagina:
        params['pagina'] = pagina
    
    url = f'{BASE_URL}/nfe/lista-notas'
    response = requests.get(url, params, headers=headers)
    
    match (response.status_code):
        case 200:
            json_data = response.json().get('dados')
            nfes = [NotaFiscal.from_json(**d) for d in json_data['itens']]
            return nfes
        case _:
            return response.text

def buscar(token_emissor: str, 
           token_secret_emissor: str, 
           *args, **kwargs):
    raise NotImplementedError

def get_nota(token_emissor: str, 
             token_secret_emissor: str, 
             *args, **kwargs):
    raise NotImplementedError

def inutilizar_numeracao(token_emissor: str, 
                         token_secret_emissor: str, 
                         token_empresa:str, 
                         token_secret_empresa:str, 
                         *args, **kwargs):
    raise NotImplementedError

def get_pre_visualizacao(token_emissor: str, 
                         token_secret_emissor: str,
                         token_empresa:str, 
                         token_secret_empresa:str, 
                         *args, **kwargs):
    raise NotImplementedError

def get_danfe(token_emissor: str, 
              token_secret_emissor: str, 
              token_empresa:str, 
              token_secret_empresa:str, 
              *args, **kwargs):
    raise NotImplementedError

# =====================================================================


''' 
    Módulo específico para geração de Notas Fiscais de Serviço.
    
    Para utilizar esse módulo basta importá-lo da seguinte forma:  
    `from pysisnoapi import nfse`
'''

# =====================================================================
from . import *

import json
import requests
from datetime import datetime
import jsonpickle

# =====================================================================
CSV_HEADERS = [
    'CNPJ',
    'CPF',
    'Nome',
    'Indicador IE',
    'Consumidor Final',
    'Faz Retenção',
    'CEP',
    'UF',
    'IBGE',
    'Bairro',
    'Endereço',
    'Número',
    'Complemento',
    'E-mail',
    'Serviço',
    'Informações Complementares',
]

INDICADORES_EXIGIBILIDADE_ISS = {
    '1': 'Exigível',
    '2': 'Não incidência',
    '3': 'Isenção',
    '4': 'Exportação',
    '5': 'Imunidade',
    '6': 'Exigibilidade suspensa por decisão judicial',
    '7': 'Exigibilidade suspensa por processo administrativo',
}

INDICADORES_INCENTIVO_FISCAL = {
    '1': 'Não',
    '2': 'Sim',
}

INDICADORES_ISS_RETIDO = {
    '1': 'Sim',
    '2': 'Não',
} # TODO: Sugerir a utilização de Boolean.

RESPONSAVEIS_RETENCAO_ISS = {
    '1': 'Tomador',
    '2': 'Intermediário',
}

STATUS = [
    'aprovado',
    'reprovado',
    'contingencia',
    'cancelado',
    'Em digitação',
]

# =====================================================================
@dataclass
class Servico:
    '''_summary_

    Returns:
        _type_: _description_
    '''
    valor_servicos            : str
    discriminacao             : str
    impostos                  : 'ImpostosServico'
    
    iss_retido                : Optional[str]         = '2'
    responsavel_retencao_iss  : Optional[str]         = '1'
    
    intermediario             : Optional['Cliente']   = None
    deducoes                  : Optional[str]         = None
    desconto_incondicionado   : Optional[str]         = None
    desconto_condicionado     : Optional[str]         = None
    outras_retencoes          : Optional[str]         = None
    informacoes_complementares: Optional[str]         = None
    data_competencia          : Optional[str]         = None
    uf_local_prestacao        : Optional['Uf']        = None
    municipio_local_prestacao : Optional['Municipio'] = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_iss_retido()
        self.validate_responsavel_retencao_iss()
    
    def validate_iss_retido(self, *args, **kwargs):
        if self.iss_retido not in INDICADORES_ISS_RETIDO:
            raise ValueError(f'Indicador de ISS retido {self.iss_retido} inválido.')
    
    def validate_responsavel_retencao_iss(self, *args, **kwargs):
        if self.responsavel_retencao_iss not in RESPONSAVEIS_RETENCAO_ISS:
            raise ValueError(f'Responsável pela retenção do ISS {self.responsavel_retencao_iss} inválido.')

@dataclass
class ConstrucaoCivil:
    codigo_obra: Optional[str] = None
    art        : Optional[str] = None
    
@dataclass
class ObjetoEmissaoNFSe(BaseClass):
    cliente: 'Cliente'
    servico: 'Servico'
    
    construcao_civil: Optional['ConstrucaoCivil'] = None

@dataclass
class NotaFiscalServico(BaseClass):
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    
    id                   : Optional[int]        = None
    empresa              : Optional[Empresa]    = None
    uuid                 : Optional[str]        = None
    modelo               : Optional[str]        = None
    status               : Optional[str]        = None
    motivo               : Optional[str]        = None
    numero_nota          : Optional[str]        = None
    codigo_verificacao   : Optional[str]        = None
    protocolo            : Optional[str]        = None
    nome_destinatario    : Optional[str]        = None
    uf_destinatario      : Optional[str]        = None
    cpf_cnpj_destinatario: Optional[str]        = None
    valor_total          : Optional[str]        = None
    data_emissao         : Optional[str]        = None
    data_competencia     : Optional[str]        = None
    uf_prestacao         : Optional[Uf]         = None
    municipio_prestacao  : Optional[Municipio]  = None
    ambiente             : Optional[str]        = None
    json_objeto_nfse     : Optional[str]        = None  # TODO: Não consta na documentação da API
    xml                  : Optional[str]        = None  # TODO: Não consta na documentação da API
    
    numer_nota           : Optional[str]        = None  # TODO: EXCLUIR! É um typo de "numero_nota", adicionei hoje (03/07/2023) apenas para o package não parar de funcionar.
    
    @classmethod
    def from_json(cls, **kwargs):
        empresa_dict = kwargs.pop('empresa', {})
        empresa = Empresa.from_json(**empresa_dict)
        
        uf_dict = kwargs.pop('uf_prestacao', {})
        uf      = Uf.from_json(**uf_dict)
        
        municipio_dict = kwargs.pop('municipio_prestacao', {})
        municipio = Municipio.from_json(**municipio_dict)
        
        return cls(empresa=empresa, uf_prestacao=uf, municipio_prestacao=municipio, **kwargs)

    def to_json(self, *args, **kwargs):
        json_str = jsonpickle.encode(self.as_filtered_dict(), unpicklable=False)
        return json_str
    
@dataclass
class PaginaNotasServico:
    total           : Optional[str]                       = None
    itens_por_pagina: Optional[str]                       = None
    pagina_atual    : Optional[str]                       = None
    itens           : Optional[List['NotaFiscalServico']] = None
    
@dataclass
class ImpostosServico(Impostos):
    issqn: 'Issqn'

@dataclass
class Issqn:
    indicador_exigibilidade_iss: str
    indicador_incentivo_fiscal : str
    item_lista_servicos        : str
    aliquota                   : str
    
    numero_processo            : Optional[str] = None
    codigo_servico             : Optional[str] = None
    aliquota_retencao          : Optional[str] = None
    aliquota_irrf              : Optional[str] = None
    aliquota_csll              : Optional[str] = None
    aliquota_previdencia_social: Optional[str] = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_indicador_exigibilidade_iss()
        self.validate_indicador_incentivo_fiscal()
    
    def validate_indicador_exigibilidade_iss(self, *args, **kwargs):
        if self.indicador_exigibilidade_iss not in INDICADORES_EXIGIBILIDADE_ISS:
            raise ValueError(f'Indicador de Exigibilidade Fiscal {self.indicador_exigibilidade_iss} inválido.')
    
    def validate_indicador_incentivo_fiscal(self, *args, **kwargs):
        if self.indicador_incentivo_fiscal not in INDICADORES_INCENTIVO_FISCAL:
            raise ValueError(f'Indicador de Incentivo Fiscal {self.indicador_incentivo_fiscal} inválido.')
    
# =====================================================================
def emitir(token_emissor: str, 
           token_secret_emissor: str,
           token_empresa: str, 
           token_secret_empresa: str, 
           objetoNfse: ObjetoEmissaoNFSe, 
           *args, **kwargs):
    '''Método responsável por enviar uma requisição para a plataforma SISNO solicitando a emissão de uma nova fiscal de SERVIÇO.
    
    Args:
        obj_emissao_nfse (ObjetoEmissaoNFSe): Objeto da classe "ObjetoEmissaoNFSe" que contém todos os dados necessários

    Returns:
        dict: Dicionário com os dados da requisição
        str: Com o response.text em caso de falha da requisição
    '''

    headers = HEADERS.copy()
    
    # -----------------------------------------
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    validate_tokens(token_empresa, token_secret_empresa)
    headers['token-empresa']        = token_empresa
    headers['token-secret-empresa'] = token_secret_empresa
    
    # -----------------------------------------
    json_str = jsonpickle.encode(objetoNfse.as_filtered_dict(), unpicklable=False)
    
    url     = f'{BASE_URL}/nfse'
    response = requests.post(url, headers=headers, json=json.loads(json_str))
    
    match (response.status_code):
        case 200:
            return response.json()
        case _:
            return response.text

def buscar_notas(token_emissor: str, 
                 token_secret_emissor: str,
                 cnpjEmpresa:str=None, 
                 data_inicio:datetime=None, 
                 data_fim:datetime=None, 
                 ambiente:str=None, 
                 status:str=None, 
                 texto:str=None, 
                 pagina:int=None, 
                 qtd_por_pagina:int=None, 
                 ordencao:str=None, 
                 tipo_ordenacao:str=None,
                 *args, **kwargs) -> List[NotaFiscalServico]:
    '''Recupera as notas fiscais de serviço.

    Args:
        cnpj (str): CNPJ Empresa (apenas números).
        
        data_inicio (datetime): Início intervalo de datas (dd/MM/yyyy HH:mm:ss).
        
        data_fim (datetime): Fim intervalo de datas (dd/MM/yyyy HH:mm:ss).
        
        ambiente (str):  
            1: Produção  
            2: Homologação
            
        status (str): Status da NFSe.
            - aprovado  
            - reprovado  
            - contingencia    
            - cancelado  
            - Em digitação
            
        texto (str): Texto de busca livre.
        
        pagina (int): Número da página.
        
        qtd_por_pagina (int): Quantidade de notas por página.
        
        ordencao (str): Campo para ordenação das notas.  
            - empresa  
            - ambiente  
            - numero_nota  
            - status  
            - data_emissao  
            - nome_destinatario  
            - uf_destinatario  
            - valor_total
            
        tipo_ordenacao (str): Tipo de Ordenação.
            - desc  
            - asc  
    
    Returns:
        List[NotaFiscalServico]: Lista com todas as NFSe
    '''
    
    # TODO: o parâmetro cnpjEmpresa está errado na documentação ('CNPJ Empresa')
    # TODO: o endpoint ignora o parâmetro cnpjEmpresa em alguns casos, descrir quais os casos.
    # TODO: textoBusca é case sensitive e leva em consideração acentos.
    # TODO: dataFim não precisa de dataInicio
    
    headers = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    params = {}
    if cnpjEmpresa:
        params['cnpjEmpresa'] = cnpjEmpresa
    if data_inicio:
        params['dataInicio'] = data_inicio.strftime('%d/%m/%Y %H:%M:%S')
    if data_fim:
        params['dataFim'] = data_fim.strftime('%d/%m/%Y %H:%M:%S')
    if ambiente:
        if ambiente not in AMBIENTES:
            raise ValueError(f'Ambiente {ambiente} inválido.')
        params['ambiente'] = ambiente      
    if status:
        if status not in STATUS:
            raise ValueError(f'Status {status} inválido.')
        params['status'] = status
    if texto:
        params['textoBusca'] = texto
    if pagina:
        params['pagina'] = str(pagina)
    if qtd_por_pagina:
        params['qtdPorPagina'] = str(qtd_por_pagina)
    if ordencao:
        params['ordenacao'] = ordencao
    if tipo_ordenacao:
        params['tipoOrdenacao'] = tipo_ordenacao

    url = f'{BASE_URL}/nfse'
    
    response  = requests.get(url, headers=headers, params=params)

    match (response.status_code):
        case 200:
            json_dados = response.json().get('dados')
            nfses      = [NotaFiscalServico.from_json(**d) for d in json_dados['itens']]
            return (response, nfses)
    
    return (response, None)

def retransmitir(token_emissor: str, 
                 token_secret_emissor: str,
                 token_empresa:str, 
                 token_secret_empresa:str, 
                 *args, **kwargs):
    raise NotImplementedError

def recuperar_dados(token_emissor: str, 
                    token_secret_emissor: str,
                    token_empresa:str, 
                    token_secret_empresa:str, 
                    id_nfse:int,
                    *args, **kwargs):
    # TODO: Cada NFSe possui um ID mesmo que de empresas diferentes ?
    # TODO: Essa função deveria estar atrelada as chaves de API, uma vez que será através delas que emitiremos as notas por uma empresa ou por outra ?

    if not isinstance(id_nfse, int):
        raise ValueError('Necessário informar um ID de NFSe válido.')
    
    headers  = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    validate_tokens(token_empresa, token_secret_empresa)
    headers['token-empresa']        = token_empresa
    headers['token-secret-empresa'] = token_secret_empresa
    
    url      = f'{BASE_URL}/nfse/{id_nfse}'
    response = requests.get(url, headers=headers)

    match (response.status_code):
        case 200:
            return response.json().get('dados')
        case 412:
            return []
        case _:
            return

def cancelar(token_emissor: str, 
             token_secret_emissor: str, 
             token_empresa:str,
             token_secret_empresa:str, 
             *args, **kwargs):
    raise NotImplementedError

# =====================================================================



'''
    Package criado para integração com a API da plataforma SISNO.

    Para utilizar esse package basta importá-lo da seguinte forma:  
    `import pysisnoapi`  
    
    Ou também poderá importar apenas os módulos necessários, como por exemplo:  
    `from pysisnoapi import nfe` 
'''

# ======================================================================================================================
# Imports:
import os
import functools

from dataclasses import dataclass, asdict
from typing      import Optional, List, Tuple, Dict
from datetime    import datetime
from dateutil    import parser

# ======================================================================================================================
# Globals:

BASE_URL = 'https://homolog.arkaonline.com.br/nfe-service'               # TODO: Provisório, em breve voltará para 'https://homolog.sisno.com.br/nfe-service/'

HEADERS  = {
    'accept'      : 'application/json',
    'Content-Type': 'application/json',
}

# -------------------------------------------
AMBIENTES = {
    '1': 'Produção',
    '2': 'Homologação',
}

FORMAS_PAGAMENTO = {
    '0': 'À Vista',
    '1': 'À Prazo',
}

MEIOS_PAGAMENTO = {
    '01': 'Dinheiro',
    '02': 'Cheque',
    '03': 'Cartão de crédito',
    '04': 'Cartão de débito',
    '05': 'Cartão da loja',
    '10': 'Vale alimentação',
    '11': 'Vale refeição',
    '12': 'Vale presente',
    '13': 'Vale combustível',
    '14': 'Duplicata mercantil',
    '15': 'Boleto bancário',
    '16': 'Depósito bancário',
    '17': 'Pagamento Instantaneo (PIX)',
    '18': 'Transferência bancária',
    '19': 'Programa de fidelidade (Cashback)',
    '90': 'Sem pagamento',
    '99': 'Outros',
}

MOTIVOS_DESONERACAO = {
    '1' : 'Táxi',
    '3' : 'Produtor Agropecuário',
    '4' : 'Frotista Locadora',
    '5' : 'Diplomático Consular',
    '6' : 'Utilitários Motocicletas Amazônia Ocidental Áreas Livre Comércio',
    '7' : 'Suframa',
    '8' : 'Venda Órgãos Públicos',
    '9' : 'Outros',
    '10': 'Deficiente Condutor',
    '11': 'Deficiente Não Condutor',
    '12': 'Órgão de Fomento Desenvolvimento Agropecuário',
    '90': 'Solicitado pelo Fisco',
}

SITUACOES_TRIBUTARIAS_ICMS = {
    '00' : 'Tributada integralmente',
    '10' : 'Tributada com cobrança de ICMS por ST',
    '20' : 'Com redução da base de cálculo',
    '30' : 'Isenta ou não tributada com cobrança de ICMS por ST',
    '40' : 'Isenta',
    '41' : 'Não tributada',
    '50' : 'Suspensão',
    '51' : 'Diferimento',
    '60' : 'ICMS cobrado anteriormente por ST',
    '70' : 'Com redução da base de cálculo/Cobrança ICMS por ST/ICMS ST',
    '90' : 'Outros',
    '101': 'Tributada pelo Simples Nacional com permissão de crédito',
    '102': 'Tributada pelo Simples Nacional sem permissão de crédito',
    '103': 'Isenção do ICMS no Simples Nacional para faixa de receita bruta',
    '201': 'Tributada pelo Simples Nacional com permissão de crédito e com cobrança do ICMS por substituição tributária',
    '202': 'Tributada pelo Simples Nacional sem permissão de crédito e com cobrança do ICMS por substituição tributária',
    '203': 'Isenção do ICMS no Simples Nacional para faixa de receita bruta e com cobrança do ICMS por substituição tributária',
    '300': 'Imune',
    '400': 'Não tributada pelo Simples Nacional',
    '500': 'ICMS cobrado anteriormente por substituição tributária (substituído) ou por antecipação',
    '900': 'Outros',
}

SITUACOES_TRIBUTARIAS_IPI = {
    '00': 'Entrada com Recuperação de Crédito',
    '01': 'Entrada Tributada com Alíquota Zero',
    '02': 'Entrada Isenta',
    '03': 'Entrada Não Tributada',
    '04': 'Entrada Imune',
    '05': 'Entrada com Suspensão',
    '49': 'Outras Entradas',
    '50': 'Saída Tributada',
    '51': 'Saída Tributável com Alíquota Zero',
    '52': 'Saída Isenta',
    '53': 'Saída Não Tributada',
    '54': 'Saída Imune',
    '55': 'Saída com Suspensão',
    '99': 'Outras Saídas',
}

SITUACOES_TRIBUTARIAS_PIS_COFINS = {
    '01': 'Operação Tributável com Alíquota Básica',
    '02': 'Operação Tributável com Alíquota por Unidade de Medida de Produto',
    '03': 'Operação Tributável com Alíquota por Unidade de Medida de Produto',
    '04': 'Operação Tributável Monofásica - Revenda a Alíquota Zero',
    '05': 'Operação Tributável por Substituição Tributária',
    '06': 'Operação Tributável a Alíquota Zero',
    '07': 'Operação Isenta da Contribuição',
    '08': 'Operação sem Incidência da Contribuição',
    '09': 'Operação com Suspensão da Contribuição',
    '49': 'Outras Operações de Saída',
    '50': 'Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Tributada no Mercado Interno',
    '51': 'Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Não-Tributada no Mercado Interno',
    '52': 'Operação com Direito a Crédito - Vinculada Exclusivamente a Receita de Exportação',
    '53': 'Operação com Direito a Crédito - Vinculada a Receitas Tributadas e Não-Tributadas no Mercado Interno',
    '54': 'Operação com Direito a Crédito - Vinculada a Receitas Tributadas no Mercado Interno e de Exportação',
    '55': 'Operação com Direito a Crédito - Vinculada a Receitas Não Tributadas Mercado Interno e de Exportação',
    '56': 'Oper. c/ Direito a Créd. Vinculada a Rec. Tributadas e Não-Tributadas Mercado Interno e de Exportação',
    '60': 'Crédito Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Tributada no Mercado Interno',
    '61': 'Créd. Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Não-Tributada Mercado Interno',
    '62': 'Crédito Presumido - Operação de Aquisição Vinculada Exclusivamente a Receita de Exportação',
    '63': 'Créd. Presumido - Oper. de Aquisição Vinculada a Rec.Tributadas e Não-Tributadas no Mercado Interno',
    '64': 'Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Tributadas no Mercado Interno e de Exportação',
    '65': 'Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Não-Tributadas Mercado Interno e Exportação',
    '66': 'Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Trib. e Não-Trib. Mercado Interno e Exportação',
    '67': 'Crédito Presumido - Outras Operações',
    '70': 'Operação de Aquisição sem Direito a Crédito',
    '71': 'Operação de Aquisição com Isenção',
    '72': 'Operação de Aquisição com Suspensão',
    '73': 'Operação de Aquisição a Alíquota Zero',
    '74': 'Operação de Aquisição sem Incidência da Contribuição',
    '75': 'Operação de Aquisição por Substituição Tributária',
    '98': 'Outras Operações de Entrada',
    '99': 'Outras Operações',
}

TIPOS_CONSUMIDOR_FINAL = {
    '0': 'Não',
    '1': 'Sim',
} # TODO: Sugerir utilização de Boolean

TIPOS_CONTRIBUINTES = {
    '1': 'Contribuinte ICMS',
    '2': 'Contribuinte isento',
    '9': 'Não contribuinte',
}

UNIDADES = [
    'AMPOLA',
    'BALDE'
    'BANDEJ',
    'BARRA',
    'BISNAG',
    'BLOCO',
    'BOBINA',
    'BOMB',
    'CAPS',
    'CART',
    'CENTO',
    'CJ',
    'CM',
    'CM2',
    'CX',
    'CX2',
    'CX3',
    'CX5',
    'CX10',
    'CX15',
    'CX20',
    'CX25',
    'CX50',
    'CX100',
    'DISP',
    'DUZIA',
    'EMBAL',
    'FARDO',
    'FOLHA',
    'FRASCO',
    'GALAO',
    'JOGO',
    'KG',
    'KIT',
    'LATA',
    'LITRO',
    'M',
    'M2',
    'M3',
    'MILHEI',
    'ML',
    'MWH',
    'PACOTE',
    'PALETE',
    'PARES',
    'PC',
    'POTE',
    'K',
    'RESMA',
    'ROLO',
    'SACO',
    'SACOLA',
    'TAMBOR',
    'TANQUE',
    'TON',
    'TUBO',
    'UNID',
    'VASIL',
    'VIDRO',
]

# ======================================================================================================================
# Classes:
@dataclass
class BaseClass:
    '''Classe genérica utilizada como base para as demais classes neste package.
    
    A principal finalidade desta classe é fornecer o método `as_filtered_dict` para filtrar e retornar um dicionário com os atributos relevantes da instância.
    Para maiores informações consulte a documentação do método [aqui](#pysisnoapi.BaseClass.as_filtered_dict).
    '''
    
    def as_filtered_dict(self)-> dict[str:str]:
        '''
        Retorna um dicionário contendo os campos/atributos da classe com valores não nulos.

        O dicionário resultante possui chaves (str) correspondentes aos nomes dos campos
        da classe e valores (str) representando os respectivos valores desses campos,
        desde que esses valores não sejam nulos (None).
        
        Returns:
            dict: Um dicionário que mapeia os campos não nulos da classe em pares chave-valor.
        '''
        return asdict(self, dict_factory=_dict_factory)

@dataclass
class Cfop:
    '''Classe `CFOP` (Código Fiscal de Operações e Prestações)
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    codigo: Optional[str] = None
    descricao: Optional[str] = None
    aplicacao: Optional[str] = None

@dataclass
class Cliente:
    '''Classe `Cliente`
    Geralmente é o destinatário da NFe.
    '''
    consumidor_final: str
    contribuinte    : str
    endereco        : 'Endereco'
    
    pessoa_fisica   : Optional['PessoaFisica']   = None
    pessoa_juridica : Optional['PessoaJuridica'] = None
    
    id              : Optional[int] = None
    ie              : Optional[str] = None
    telefone        : Optional[str] = None
    email           : Optional[str] = None
    faz_retencao    : Optional[str] = None
        
    def __post_init__(self):        
        self.validate_pessoa()
        self.validate_consumidor_final()
        self.validate_contribuinte()
    
    def validate_pessoa(self):
        if self.pessoa_fisica and self.pessoa_juridica:
            raise ValueError('Informe apenas um dos campos pessoa_fisica ou pessoa_juridica')
        if not self.pessoa_fisica and not self.pessoa_juridica:
            raise ValueError('Informe pelo menos um dos campos pessoa_fisica ou pessoa_juridica')
    
    def validate_consumidor_final(self):
        if self.consumidor_final not in TIPOS_CONSUMIDOR_FINAL:
            raise ValueError(f'Consumidor Final {self.consumidor_final} inválido')
    
    def validate_contribuinte(self):
        if self.contribuinte not in TIPOS_CONTRIBUINTES:
            raise ValueError(f'Contribuinte {self.contribuinte} inválido')

@dataclass
class Cofins:
    '''Contribuição para Financiamento da Seguridade Social.

    situacao_tributaria:  
        01: Operação Tributável com Alíquota Básica  
        02: Operação Tributável com Alíquota por Unidade de Medida de Produto  
        03: Operação Tributável com Alíquota por Unidade de Medida de Produto  
        04: Operação Tributável Monofásica - Revenda a Alíquota Zero  
        05: Operação Tributável por Substituição Tributária  
        06: Operação Tributável a Alíquota Zero  
        07: Operação Isenta da Contribuição  
        08: Operação sem Incidência da Contribuição  
        09: Operação com Suspensão da Contribuição  
        49: Outras Operações de Saída  
        50: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Tributada no Mercado Interno  
        51: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Não-Tributada no Mercado Interno  
        52: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita de Exportação  
        53: Operação com Direito a Crédito - Vinculada a Receitas Tributadas e Não-Tributadas no Mercado Interno  
        54: Operação com Direito a Crédito - Vinculada a Receitas Tributadas no Mercado Interno e de Exportação  
        55: Operação com Direito a Crédito - Vinculada a Receitas Não Tributadas Mercado Interno e de Exportação  
        56: Oper. c/ Direito a Créd. Vinculada a Rec. Tributadas e Não-Tributadas Mercado Interno e de Exportação  
        60: Crédito Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Tributada no Mercado Interno  
        61: Créd. Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Não-Tributada Mercado Interno  
        62: Crédito Presumido - Operação de Aquisição Vinculada Exclusivamente a Receita de Exportação  
        63: Créd. Presumido - Oper. de Aquisição Vinculada a Rec.Tributadas e Não-Tributadas no Mercado Interno  
        64: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Tributadas no Mercado Interno e de Exportação  
        65: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Não-Tributadas Mercado Interno e Exportação  
        66: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Trib. e Não-Trib. Mercado Interno e Exportação  
        67: Crédito Presumido - Outras Operações  
        70: Operação de Aquisição sem Direito a Crédito  
        71: Operação de Aquisição com Isenção  
        72: Operação de Aquisição com Suspensão  
        73: Operação de Aquisição a Alíquota Zero  
        74: Operação de Aquisição sem Incidência da Contribuição  
        75: Operação de Aquisição por Substituição Tributária  
        98: Outras Operações de Entrada  
        99: Outras Operações  
    
    aliquota (str): $0.0000  
        Passar percentual quando tributado pela alíquota ou em reais quando tributado por quantidade  
    
    aliquota_st	(str): $0.0000
    
    aliquota_retencao (str): $0.0000
    '''
    situacao_tributaria : str
    
    aliquota            : Optional[str] = None
    aliquota_st         : Optional[str] = None
    aliquota_retencao   : Optional[str] = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_situacao_tributaria()
        
    def validate_situacao_tributaria(self, *args, **kwargs):
        if self.situacao_tributaria not in SITUACOES_TRIBUTARIAS_PIS_COFINS:
            raise ValueError(f'Situação tributária {self.situacao_tributaria} inválida.')

@dataclass
class DeclaracaoImportacaoAdicao:
    numero_sequencial: Optional[str] = None
    numero           : Optional[str] = None
    cod_fabricante   : Optional[str] = None
    desconto         : Optional[str] = None
    drawback         : Optional[str] = None
    
@dataclass
class Empresa:  
    '''
        Classe que irá representar todas as empresas cadastradas da plataforma SISNO.
        Pela documentação do dia 11/05/2023, essa classe basicamente só terá sua utilidade ao consultar notas.
        Em resumo: não há necessidade de se preocupar com os campos, caso não saiba algum deles, pois, eles serão preenchidos automaticamente com aquilo que vier do endpoint.
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    id                                          : Optional[str]         = None
    token                                       : Optional[str]         = None # Necessário para emitir, mas não para consultar
    token_secret                                : Optional[str]         = None # Necessário para emitir, mas não para consultar
    cnpj                                        : Optional[str]         = None
    nome_fantasia                               : Optional[str]         = None
    razao_social                                : Optional[str]         = None
    endereco                                    : Optional['Endereco']  = None
    telefone                                    : Optional[str]         = None
    inscricao_estadual                          : Optional[str]         = None
    inscricao_municipal                         : Optional[str]         = None
    inscricao_estadual_substituicao_tributaria  : Optional[str]         = None
    regime_tributario                           : Optional[str]         = None
    classificacao_nacional_atividades_economicas: Optional[str]         = None
    ambiente                                    : Optional[str]         = None
    id_csc                                      : Optional[str]         = None
    csc                                         : Optional[str]         = None
    codigo_regime_especial_tributacao           : Optional[str]         = None
    porcentagem_icms_aproveitado                : Optional[str]         = None
    site                                        : Optional[str]         = None
    email                                       : Optional[str]         = None
    utiliza_tributos_aproximados                : Optional[bool]        = None  # TODO: Não consta na documentação da API
    informacoes_complementares                  : Optional[str]         = None  # TODO: Não consta na documentação da API
    senha_portal_prefeitura                     : Optional[str]         = None  # TODO: Não consta na documentação da API
    serie_rps                                   : Optional[str]         = None  # TODO: Não consta na documentação da API 
    proximo_numero_rps                          : Optional[str]         = None  # TODO: Não consta na documentação da API 
    proximo_numero_rps_homologacao              : Optional[str]         = None  # TODO: Não consta na documentação da API 
    numero_lote_rps                             : Optional[str]         = None  # TODO: Não consta na documentação da API 
    
    @classmethod
    def from_json(cls, **kwargs):
        endereco_dict = kwargs.pop('endereco', {})
        endereco = Endereco.from_json(**endereco_dict)
        
        return cls(endereco=endereco, **kwargs)

@dataclass
class Endereco:
    '''Classe `Endereço`
    '''
    
    codigo_pais        : str
    descricao_pais     : str
    bairro             : str
    logradouro         : str
    numero             : str
    
    id                 : Optional[int] = None
    uf                 : Optional[str] = None
    codigo_municipio   : Optional[str] = None
    descricao_municipio: Optional[str] = None
    cep                : Optional[str] = None
    complemento        : Optional[str] = None
    
    @classmethod
    def from_json(cls, **kwargs):
        return cls(**kwargs)
    
@dataclass
class Ibpt:
    '''Classe `IBPT` (Impostos sobre Produtos e Serviços)
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    codigo            : Optional[str]  = None
    ex                : Optional[str]  = None
    tipo              : Optional[str]  = None
    descricao         : Optional[str]  = None
    nacional_federal  : Optional[str]  = None
    importados_federal: Optional[str]  = None
    estadual          : Optional[str]  = None
    municipal         : Optional[str]  = None
    vigencia_inicio   : Optional[str]  = None
    vigencia_fim      : Optional[str]  = None
    versao            : Optional[str]  = None
    fonte             : Optional[str]  = None
    unidade_federativa: Optional['Uf'] = None
    ativo             : Optional[str]  = None     # TODO: Até o dia 01/06/2023, não consta na Documentação
    
    @classmethod
    def from_json(cls, **kwargs):
        uf_dict = kwargs.pop('unidade_federativa', {})
        uf = Uf.from_json(**uf_dict)
        return cls(unidade_federativa=uf, **kwargs)

@dataclass
class Icms:
    '''Classe `ICMS` (Imposto sobre Circulação de Mercadorias e Serviços)
    '''
    situacao_tributaria                       : str
    
    codigo_cfop                               : Optional[str]  = None
    aliquota_icms                             : Optional[str]  = None
    percentual_reducao_bc_icms                : Optional[str]  = None
    aliquota_fcp                              : Optional[str]  = None
    percentual_reducao_bc_icms_st             : Optional[str]  = None
    aliquota_aplicavel_calculo_credito        : Optional[str]  = None
    aliquota_icms_st                          : Optional[str]  = None
    aliquota_fcp_st                           : Optional[str]  = None
    percentual_margem_valor_agregado_icms_st  : Optional[str]  = None
    percentual_diferimento                    : Optional[str]  = None
    percentual_desonerado                     : Optional[str]  = None
    motivo_desoneracao                        : Optional[str]  = None
    percentual_icms_st_retido                 : Optional[str]  = None
    utilizar_tabela_aliquotas_interestaduais  : Optional[bool] = None
    utilizar_aliquota_interestadual_importacao: Optional[bool] = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_situacao_tributaria()
        self.validate_motivo_desoneracao()
        
    def validate_situacao_tributaria(self, *args, **kwargs):
        if self.situacao_tributaria not in SITUACOES_TRIBUTARIAS_ICMS:
            raise ValueError(f'Situação tributária {self.situacao_tributaria} inválida.')
    
    def validate_motivo_desoneracao(self, *args, **kwargs):
        if self.motivo_desoneracao:
            if self.motivo_desoneracao not in MOTIVOS_DESONERACAO:
                raise ValueError(f'Motivo de Desoneração {self.motivo_desoneracao} inválido.')

@dataclass
class Impostos:
    '''Classe Base para as Classes de Impostos de Produtos e Serviços.
    '''
    pis   : 'Pis'
    cofins: 'Cofins'

@dataclass
class Ipi:
    '''Classe `IPI` (Imposto Sobre Produtos Industrializados)

    situacao_tributaria:  
        00: Entrada com Recuperação de Crédito  
        01: Entrada Tributada com Alíquota Zero  
        02: Entrada Isenta  
        03: Entrada Não Tributada  
        04: Entrada Imune  
        05: Entrada com Suspensão  
        49: Outras Entradas  
        50: Saída Tributada  
        51: Saída Tributável com Alíquota Zero  
        52: Saída Isenta  
        53: Saída Não Tributada  
        54: Saída Imune  
        55: Saída com Suspensão  
        99: Outras Saídas  
    '''
    situacao_tributaria : str
    
    aliquota            : Optional[str] = None
    codigo_enquadramento: Optional[str] = None
    codigo_selo         : Optional[str] = None
    qtd_selo            : Optional[str] = None
    
    def __post_init__(self, *args, **kwargs):
        self.validate_situacao_tributaria()
    
    def validate_situacao_tributaria(self, *args, **kwargs):
        if self.situacao_tributaria not in SITUACOES_TRIBUTARIAS_IPI:
            raise ValueError(f'Situação tributária {self.situacao_tributaria} inválida.')

@dataclass
class Municipio:
    '''Classe `Município`
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    codigo_ibge : Optional[int] = None
    descricao   : Optional[str] = None
    
    @classmethod
    def from_json(cls, **kwargs):
        return cls(**kwargs)

@dataclass
class NotaFiscal:
    '''Classe `Nota Fiscal`.
    
    Até 01/2023 essa classe também era usada para Notas Fiscais de Serviço (NFSe).
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    id                      : Optional[int]         = None
    empresa                 : Optional['Empresa']   = None
    tipo                    : Optional[str]         = None
    serie                   : Optional[str]         = None
    numero_nota             : Optional[str]         = None
    chave_acesso            : Optional[str]         = None
    protocolo               : Optional[str]         = None
    nome_destinatario       : Optional[str]         = None
    uf_destinatario         : Optional[str]         = None
    cpf_cnpj_destinatario   : Optional[str]         = None
    valor_total             : Optional[str]         = None
    status                  : Optional[str]         = None
    motivo                  : Optional[str]         = None
    data_emissao            : Optional[datetime]    = None
    data_autorizacao        : Optional[datetime]    = None
    modelo                  : Optional[str]         = None
    ambiente                : Optional[str]         = None
    xml                     : Optional[str]         = None
    json_objeto_nfe         : Optional[str]         = None
    tipo_emissao            : Optional[str]         = None
    numero_lote             : Optional[str]         = None
    
    @classmethod
    def from_json(cls, **kwargs):
        empresa_dict = kwargs.pop('empresa', {})
        empresa = Empresa.from_json(**empresa_dict)
        
        if data_emissao := kwargs.pop('data_emissao', None):
            data_emissao = parser.parse(data_emissao, dayfirst=True)
        
        if data_autorizacao := kwargs.pop('data_autorizacao', None):
            data_autorizacao = parser.parse(data_autorizacao, dayfirst=True)
        
        return cls(
            empresa=empresa,
            data_emissao=data_emissao, 
            data_autorizacao=data_autorizacao, 
            **kwargs
        )

@dataclass
class Observacao:
    campo: Optional[str] = None
    texto: Optional[str] = None
    
@dataclass
class PessoaFisica:
    '''Classe `Pessoa Física`
    
    Deve ser informado apenas um dos campos [cpf, id_estrangeiro]
    '''
    nome_completo: str
    
    cpf           : Optional[str] = None
    id_estrangeiro: Optional[str] = None
    
    def __post_init__(self):
        if self.cpf and self.id_estrangeiro:
            raise ValueError('Informe somente um dos campos: cpf ou id_estrangeiro')
        if not self.cpf and not self.id_estrangeiro:
            raise ValueError('Informe um dos campos: cpf ou id_estrangeiro')

@dataclass
class PessoaJuridica:
    '''Classe `Pessoa Jurídica`
    '''
    cnpj        : str
    razao_social: str
    
    im          : Optional[str] = None
    suframa     : Optional[str] = None

@dataclass
class Pis:
    '''Classe `PIS` (Programas de Integração Social)

    situacao_tributaria (str): Situação Tributária  
        01: Operação Tributável com Alíquota Básica  
        02: Operação Tributável com Alíquota por Unidade de Medida de Produto  
        03: Operação Tributável com Alíquota por Unidade de Medida de Produto  
        04: Operação Tributável Monofásica - Revenda a Alíquota Zero  
        05: Operação Tributável por Substituição Tributária  
        06: Operação Tributável a Alíquota Zero  
        07: Operação Isenta da Contribuição  
        08: Operação sem Incidência da Contribuição  
        09: Operação com Suspensão da Contribuição  
        49: Outras Operações de Saída  
        50: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Tributada no Mercado Interno  
        51: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita Não-Tributada no Mercado Interno  
        52: Operação com Direito a Crédito - Vinculada Exclusivamente a Receita de Exportação  
        53: Operação com Direito a Crédito - Vinculada a Receitas Tributadas e Não-Tributadas no Mercado Interno  
        54: Operação com Direito a Crédito - Vinculada a Receitas Tributadas no Mercado Interno e de Exportação  
        55: Operação com Direito a Crédito - Vinculada a Receitas Não Tributadas Mercado Interno e de Exportação  
        56: Oper. c/ Direito a Créd. Vinculada a Rec. Tributadas e Não-Tributadas Mercado Interno e de Exportação  
        60: Crédito Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Tributada no Mercado Interno  
        61: Créd. Presumido - Oper. de Aquisição Vinculada Exclusivamente a Rec. Não-Tributada Mercado Interno  
        62: Crédito Presumido - Operação de Aquisição Vinculada Exclusivamente a Receita de Exportação  
        63: Créd. Presumido - Oper. de Aquisição Vinculada a Rec.Tributadas e Não-Tributadas no Mercado Interno  
        64: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Tributadas no Mercado Interno e de Exportação  
        65: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Não-Tributadas Mercado Interno e Exportação  
        66: Créd. Presumido - Oper. de Aquisição Vinculada a Rec. Trib. e Não-Trib. Mercado Interno e Exportação  
        67: Crédito Presumido - Outras Operações  
        70: Operação de Aquisição sem Direito a Crédito  
        71: Operação de Aquisição com Isenção  
        72: Operação de Aquisição com Suspensão  
        73: Operação de Aquisição a Alíquota Zero  
        74: Operação de Aquisição sem Incidência da Contribuição  
        75: Operação de Aquisição por Substituição Tributária  
        98: Outras Operações de Entrada  
        99: Outras Operações  
    
    aliquota (str): $0.0000 
        Passar percentual quando tributado pela alíquota ou em reais quando tributado por quantidade  
    
    aliquota_st	(str): $0.0000  
    
    aliquota_retencao (str): $0.0000
    
    '''
    situacao_tributaria: str
    
    aliquota           : Optional[str] = None
    aliquota_st        : Optional[str] = None
    aliquota_retencao  : Optional[str] = None

    def __post_init__(self, *args, **kwargs):
        self.validate_situacao_tributaria()
        
    def validate_situacao_tributaria(self, *args, **kwargs):
        if self.situacao_tributaria not in SITUACOES_TRIBUTARIAS_PIS_COFINS:
            raise ValueError(f'Situação tributária {self.situacao_tributaria} inválida.')

@dataclass
class RetencaoIcmsTransporte:
    valor_servico                                           : Optional[str] = None
    valor_icms_retido                                       : Optional[str] = None
    cfop                                                    : Optional[str] = None
    codigo_municipio_ocorrencia_fato_gerador_icms_transporte: Optional[str] = None
    base_calculo_retencao_icms                              : Optional[str] = None
    aliquota_retencao                                       : Optional[str] = None
    
@dataclass
class Transporte:
    # TODO: Até o dia 10/05/2023 essa classe não está sendo usada
    ...

@dataclass
class Uf:
    '''Classe `UF`(Unidade Federativa)

    Returns:
        _type_: _description_
    '''
    # TODO: Até o dia 01/06/2023, não consta na Documentação quais são os campos obrigatórios
    # TODO: Até o dia 01/06/2023 essa classe só está sendo usada em classes de NFSe, entretanto, já deixarei ela por aqui mesmo, acredito que em breve as classes de NFe também usarão
    codigo_ibge: Optional[str]  = None
    sigla      : Optional[str]  = None
    descricao  : Optional[str]  = None

    @classmethod
    def from_json(cls, **kwargs):
        return cls(**kwargs)

# ======================================================================================================================
def validate_tokens(token:str, token_secret:str):
    '''Verifica se os tokens informados pelo usuário estão corretos e válidos.'''
    
    if not isinstance(token, str) or len(token) < 1:
        raise Exception('Token Empresa inválido')
    if not isinstance(token_secret, str) or len(token_secret) < 1:
        raise Exception('Token Secret Empresa inválido')
    
    return True

def _dict_factory(x: List[Tuple]) -> Optional[Dict]:
    '''Cria um dicionário filtrado contendo apenas as chaves cujos valores são diferentes de None.

    Essa função é chamada automaticamente durante a conversão de uma classe para um dicionário usando `dataclasses.asdict`, 
    quando é atribuída a opção `dict_factory` como seu valor. 
    
    No caso, estamos usando essa função no método `as_filtered_dict` da classe `BaseClass`.
    
    A função recebe uma lista de tuplas contendo pares chave-valor e retorna um dicionário contendo apenas as chaves cujos valores são diferentes de None.

    Args:
        x (List[Tuple]): Uma lista de tuplas contendo pares chave-valor.

    Returns:
        dict ou None: Um dicionário filtrado com as chaves cujos valores são diferentes de None ou retorna None se o dicionário resultante estiver vazio.
    '''
    dic = {k: v for (k, v) in x if v is not None}
    return dic if len(dic) > 0 else None

# ======================================================================================================================



''' 
    O módulo misc.py contém funções genéricas que não se enquadram em outros módulos.
    
    Um exemplo é a função para realizar uma requisição à API a fim de consultar informações sobre municípios. 
    Este módulo faz parte de um pacote e depende completamente de outros módulos.

    Para utilizar as funções deste módulo, basta importá-lo da seguinte forma:  
    `from pysisnoapi import misc`
'''

# ======================================================================================================================
from . import *

import requests
from typing import List

# ======================================================================================================================
def get_municipios(token_emissor: str, 
                   token_secret_emissor: str, 
                   uf: str, 
                   *args, **kwargs) -> List[Municipio]:
    '''Consulta os municípios de um determinado estado através de uma requisição à API.

    Essa função permite consultar os municípios de um estado específico através de uma requisição à API.
    Ela retorna uma lista de objetos Municipio contendo informações sobre cada município.

    Args:
        uf (str): UF do estado para o qual deseja-se consultar os municípios.

    Returns:
        List[Municipio]: Uma lista de objetos Municipio, representando os municípios do estado consultado.
    '''

    headers  = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    url      = f'{BASE_URL}/unidades-federativas/{uf}/municipios'
    response = requests.get(url, headers=headers)

    match (response.status_code):
        case 200:
            json_data  = response.json().get('dados')
            municipios = [Municipio(**d) for d in json_data]
            return municipios
        case 412:
            return []
        case _:
            return

def get_cfops(token_emissor: str, 
              token_secret_emissor: str, 
              *args, **kwargs) -> List[Cfop]:
    '''Obtém a lista de todos os CFOPs disponíveis através de uma requisição à API.

    Essa função permite obter a lista completa de CFOPs (Código Fiscal de Operações e Prestações) disponíveis através de uma requisição à API.

    Returns:
        List[Cfop]: Uma lista de objetos Cfop, representando os CFOPs disponíveis.
    '''

    headers  = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    url      = f'{BASE_URL}/cfops'
    response = requests.get(url, headers=headers)

    match (response.status_code):
        case 200:
            json_data = response.json().get('dados')
            cfops     = [Cfop(**d) for d in json_data]
            return cfops
        case 412:
            return []
        case _:
            return

def get_ibpts(token_emissor: str, 
              token_secret_emissor: str, 
              cod_desc: str, 
              uf: str, 
              *args, **kwargs) -> List[Ibpt]:
    '''
    Obtém os IBPTs através de uma requisição à API.

    Essa função permite obter os dados de IBPTs (Impostos sobre Produtos e Serviços) para um determinado item ou código, em uma determinada UF (Unidade Federativa)
    
    Args:
        cod_desc (str): Breve descrição do item ou código. Deve ter no mínimo 4 caracteres.
        uf (str): UF do estado para o qual deseja-se consultar os IBPTs.

    Raises:
        Exception: Caso o parâmetro cod_desc tenha menos de 4 caracteres.

    Returns:
        List[Ibpt]: Uma lista de objetos Ibpt representando os IBPTs.
    '''

    if len(cod_desc) < 4:
        raise Exception('Código ou Descrição precisa ter no mínimo 4 caracteres')
    
    headers  = HEADERS.copy()
    
    validate_tokens(token_emissor, token_secret_emissor)
    headers['token-emissor']        = token_emissor
    headers['token-secret-emissor'] = token_secret_emissor
    
    headers['codigo-ou-descricao'] = cod_desc
    headers['uf'] = uf

    url      = f'{BASE_URL}/ibpts'
    response = requests.get(url, headers=headers)

    match (response.status_code):
        case 200:
            json_data = response.json().get('dados')
            ibpts     = [Ibpt.from_json(**d) for d in json_data]
            return ibpts
        case 412:
            return []
        case _:
            return

# ======================================================================================================================
